/*
 *  (C) Copyright 2003 Pawel Parys <pp209216@zodiac.mimuw.edu.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string>
#include <map>
#include <errno.h>
#include <dirent.h>
#include <sys/sem.h>

char dirname[] = "/var/spool/ppmail";
char logfile[] = "/var/log/ppmail.log";
char sentlogfile[] = "/var/log/ppmailsent.log";
#define cfgfile "/etc/ppmail"

using namespace std;

int sock;
FILE *log;

//////////////////// NETWORK FUNCTIONS ////////////////////////////

void send(string text)
{
	int r = write(sock, text.c_str(), text.size());
	fprintf(log, ">%s", text.c_str()); fflush(log);
	if (r<0) { fprintf(log, "Error writing: %s\n", strerror(errno)); exit(1); }
}

string recv()
{
	string buf;
	int r = 0;
	for (;;)
	{
		char ch;
		r += read(sock, &ch, 1);
		buf += ch;
		if (r<0) { fprintf(log, "Error reading: %s\n", strerror(errno)); exit(1); }
		if (r==0 || ch=='\n') break;
	}
	fprintf(log, "<%s", buf.c_str()); fflush(log);
	return buf;
}

void connect_with(string target, int port_no)
{
	int colon = target.find(':');
	if (colon>=0)
	{
		sscanf(target.substr(colon+1).c_str(), "%d", &port_no);
		target = target.substr(0, colon);
	}
	printf("Connecting with %s, port = %d\n", target.c_str(), port_no); fflush(stdout);

	sockaddr_in adr;

	adr.sin_family = AF_INET;
	adr.sin_port = htons(port_no);

	int a = inet_addr(target.c_str());
	if (a==INADDR_NONE)
	{
                hostent *host = gethostbyname(target.c_str());
		if (host==NULL) throw 0;
		memcpy(&adr.sin_addr, host->h_addr, host->h_length);
	}
	else
		adr.sin_addr.s_addr = a;

	protoent *proto = getprotobyname("tcp");
	sock = socket(PF_INET, SOCK_STREAM, proto->p_proto);

	if (connect(sock, (sockaddr*)&adr, sizeof(adr))<0)
		throw 0;
}

/////////////////// SEMAPHORE HANDLING ////////////////////////

int semid;

void create_semaphore()
{
	semid = semget(IPC_PRIVATE, 1, 0600);
	semctl(semid, 0, SETVAL, 0);
}

void semaphore_move(int dir)
{
	sembuf buf;
	buf.sem_num = 0;
	buf.sem_op = dir;
	buf.sem_flg = 0;
	semop(semid, &buf, 1);
}

/////////////////// DEAMON - WAITING FOR MESSAGES //////////////////////

void connection()
// RECEIVING MESSAGE
{
	fprintf(log, "Connection (socket=%d)\n", sock); fflush(log);
	send("220 localhost PP ESMTP\r\n");
	recv();
	send("250 localhost\r\n");
	char name[1000], name2[1000];
	sprintf(name, "%s/tmp-%d-%d", dirname, (int)time(NULL), (int)random());
	sprintf(name2, "%s/mail-%d-%d", dirname, (int)time(NULL), (int)random());
	fprintf(log, "Saving to file %s\n", name2); fflush(log);
	FILE *f = fopen(name, "w");
	if (f==NULL) { fprintf(log, "Error creating file\n"); exit(1); }
	for(;;)
	{
		string str = recv();
		fprintf(f, str.c_str());
		if (str=="DATA\r\n")
		{
			send("354 go ahead\r\n");
			for (;;)
			{
				string str = recv();
				fprintf(f, str.c_str());
				if (str==".\r\n") break;
			}
			send("250 ok\r\n");
			recv(); //QUIT
			break;
		}
		send("250 ok\r\n");
	}
	fclose(f);
	rename(name, name2);
	close(sock);
	semaphore_move(1);
}

void sending_deamon()
// SEND MESSAGES LOOP
{
	stdout = fopen(sentlogfile, "a");
	for(;;)
	{
		int send_waiting();
		if (!send_waiting())
			semaphore_move(-1);
		else
			sleep(10);
	}
}

int port_no = 25;
int port_deamon = 6667;

void deamon()
// INITIALIZE DEAMON
{
	create_semaphore();
	if (fork())
		sending_deamon();
	log = fopen(logfile, "a");
	umask(0077);
	char host_name[1000], host_addr[1000];
	gethostname(host_name, sizeof(host_name));
	hostent *host_info = gethostbyname(host_name);
	char *buf2 = inet_ntoa(*((in_addr*)host_info->h_addr));
	strcpy(host_addr, buf2);
	protoent *proto = getprotobyname("tcp");
	int globalsock = socket(PF_INET, SOCK_STREAM, proto->p_proto);
	sockaddr_in adr;
	adr.sin_family = AF_INET;
	adr.sin_addr.s_addr = inet_addr(host_addr);
	adr.sin_port = htons(port_deamon);
	int e=0;
	e|=bind(globalsock, (sockaddr*)&adr, sizeof(adr));
	if (e) { fprintf(log, "Error0: %s\n", strerror(errno)); exit(1); }
	e|=listen(globalsock, SOMAXCONN);
	if (e) { fprintf(log, "Error: %s\n", strerror(errno)); exit(1); }
	for (;;)
	{
		fprintf(log, "Waiting for connection\n"); fflush(log);
		sock = accept(globalsock, NULL, 0);
		connection();
//		close(globalsock);
//		break;
	}
}


//////////////////////// CONFIGURATION FILE  /////////////////////////////////

struct AdrStruct
{
	string address, server, user, password;
	AdrStruct(const string &a, const string &s, const string &u, const string &p)
		: address(a), server(s), user(u), password(p) {}
};

map<string, AdrStruct> adrMap;
int line;

void error()
{
	printf("Error in configuration file, line %d\n", line);  fflush(stdout);
	exit(1);
}

int isSpace(char ch)
{
	return ch==' ' || ch=='\t' || ch=='\n' || ch=='\r';
}

void skipSpaces(const string &str, int &pos)
{
	while (pos<str.length() && isSpace(str[pos])) pos++;
	if (pos<str.length() && str[pos]=='#') pos = str.length();
}

int readStr(const string &str, int &pos, string &res)
{
	res="";
	skipSpaces(str, pos);
	if (pos==str.length()) return 0;
	if (str[pos]=='"')
	{
		pos++;
		for (;;)
		{
			if (pos==str.length()) error();
			if (str[pos]=='"' && (pos+1==str.length() || str[pos+1]!='"'))
				break;
			if (str[pos]=='"') pos++;
			res += str[pos++];
		}
		pos++;
	}
	else
	{
		while (pos<str.length() && !isSpace(str[pos]) && str[pos]!='#')
		{
			if (str[pos]=='"') error();
			res += str[pos++];
		}
	}
	return 1;
}

void read_config()
{
	FILE *f = fopen(cfgfile, "r");
	if (!f) { printf("Can not open \""cfgfile"\"!\n");  fflush(stdout); exit(1); }
	char buf[2000];
	string name, address, server, user, password;
	line = 1;
	for (;;)
	{
		char *r = fgets(buf, sizeof(buf), f);
		if (r==NULL) break;
		string str = buf;
		int pos = 0;
		int res0 = readStr(str, pos, name);
		int res1 = readStr(str, pos, address);
		res1 &= readStr(str, pos, server);
		int res2 = readStr(str, pos, user);
		int res3 = readStr(str, pos, password);
		skipSpaces(str, pos);
		if ((res0 && !res1) || (res2 && !res3) || pos!=str.length())
			error();
		if (res0)
			adrMap.insert(pair<string, AdrStruct>(name, AdrStruct(address, server, user, password)));
		line++;
	}
	fclose(f);
}

/////////////////////////////// SENDING (TEXT FUNCTIONS) ////////////////////////////////

void show(char *name)
{
	printf("%s\n", name); fflush(stdout);
}

string extract(const string &str, char beg, char end)
{
	int pos = 0;
	while (str.at(pos)!=beg) pos++;
	pos++;
	int pos2 = pos;
	while (str.at(pos2)!=end) pos2++;
	return str.substr(pos, pos2-pos);
}

string tobase64(const string &txt)
{
        unsigned char code[100]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
        unsigned char w[10];
        string out;
        for (int i=0; i<txt.length(); i+=3)
        {
        	int ii = out.size();
        	out.resize(ii+4);
                w[0] = txt[i]>>2;
                w[0] = code[w[0]];
                w[1] = ((unsigned char)(txt[i]<<6)>>2)|(txt[i+1]>>4);
                w[1] = code[w[1]];
                if (i+2<txt.length())
                {
                        w[2] = code[(unsigned char)(((unsigned char)(txt[i+1]<<4)>>2)|(txt[i+2]>>6))];
                        w[3] = code[(unsigned char)((unsigned char)(txt[i+2]<<2)>>2)];
                        out[ii]=w[0]; out[ii+1]=w[1];
                        out[ii+2]=w[2]; out[ii+3]=w[3];
                }
                else if(i+1<txt.length())
                {
                        w[2] = code[(unsigned char)(((unsigned char)(txt[i+1]<<4)>>2)|(txt[i+2]>>6))];
                        out[ii]=w[0]; out[ii+1]=w[1];
                        out[ii+2]=w[2]; out[ii+3]='=';
                }
                else
                {
                        out[ii]=w[0]; out[ii+1]=w[1];
                        out[ii+2]='='; out[ii+3]='=';
                }
        }
        return out;
}

////////////////////////////////// SENDING MESSAGE ////////////////////////////////

void send_error_message(const string &from, const string &text)
{
try{
	connect_with("localhost", 25);
	send("HELO ppmail\r\n");
	send("MAIL FROM: ppmail@localhost\r\n");
	send("RCPT TO: "+from+"@localhost\r\n");
	send("DATA\r\n");
	send("Subject: Error while sending\r\n");
	send("\r\n");
	send(text+"\r\n.\r\n"); 
//	for (int a=0; a<6; a++) recv();
	send("QUIT\r\n");
	close(sock);
}catch(...){}
}

void sendmes(char *name)
{
	printf("Sending "); fflush(stdout);
	show(name);
	char buf[2000], fname[2000];
	sprintf(fname, "%s/%s", dirname, name);
	FILE *f = fopen(fname, "r");
	if (!f) { printf("Error opening file: %s\n", strerror(errno)); fflush(stdout); exit(1); }
	sock = -1;
	try
	{
		for (;;)
		{
			char *r = fgets(buf, sizeof(buf), f);
			if (r==NULL) throw 0;
			if (memcmp(buf, "MAIL FROM:", 10)==0)
				break;
		}
		fseek(f, 0, SEEK_SET);
		map<string, AdrStruct>::const_iterator it = adrMap.find(extract(buf, '<', '@'));
		if (it==adrMap.end()) throw 0;
		const AdrStruct &adr = it->second;
		printf("%s %s\n", tobase64(adr.user).c_str(), tobase64(adr.password).c_str()); fflush(stdout);
		connect_with(adr.server, 25);
		recv();
		char hostname[1000]; 
		gethostname(hostname, sizeof(hostname));
		string he = adr.user.empty() ? "HE" : "EH";
		send(he+"LO "+hostname+"\r\n");
		for (;;)
		{
			string res = recv();
			if (res.size()>=4 && res[3]==' ') break;
		}
		int subst = 0, okcount = 0, data = 0;
		if (!adr.user.empty())
		{
			send("AUTH LOGIN\r\n");
			recv();
			send(tobase64(adr.user)+"\r\n");
			recv();
			send(tobase64(adr.password)+"\r\n");
			string res = recv();
			if (res.length()<4 || res.substr(0, 4)!="235 ")
				throw 0; 
		}
		for (;;)
		{
			char *r = fgets(buf, sizeof(buf), f);
			if (r==NULL || r[0]==8) break;
			if (subst<2 && memcmp(buf, "MAIL FROM:", 10)==0)
			{
				subst++;
				string a;
				try
				{
					a = extract(adr.address, '<', '>');
				}
				catch (...)
				{
					a = adr.address;
				}
				sprintf(buf, "MAIL FROM:<%s>\r\n", a.c_str());
			}
			if (subst<2 && memcmp(buf, "From:", 5)==0)
			{
				subst++;
				sprintf(buf, "From: %s\r\n", adr.address.c_str());
			}
			if (!data &&strcmp(buf, "DATA\r\n")==0) data=1;
			if (!data) okcount++;
			send(buf);
		}
		int error = 0;
		string error_text;
		for (int a = 0; a<okcount+2; a++)
		{
			string res = recv();
			if (a<okcount && (res.length()<4 || res.substr(0, 4)!="250 "))
			{
				error = 1;
				error_text += res;
			}
			if ((a>okcount && (res.length()<4 || res.substr(0, 4)!="250 ")) 
			 || (a==okcount && (res.length()<4 || res.substr(0, 4)!="354 ")))
				throw 0;
		}
		send("QUIT\r\n");
		fclose(f);
		close(sock);
		if (error)
			send_error_message(it->first, error_text);
		unlink(fname);
	}
	catch (...)
	{
		if (sock!=-1) close(sock);
		fclose(f);
		printf("Error while sending\n"); fflush(stdout);
	}
}

//////////////////////////////// MAIN //////////////////////////////////////

int for_every(void (*fun)(char*))
{
	DIR *dir = opendir(dirname);
	int count = 0;
	for (;;)
	{
		dirent *res = readdir(dir);
		if (res == NULL)
			break;
		if (memcmp(res->d_name, "mail-", 5)!=0) 
			continue;
		fun(res->d_name);
		++count;
	}
	closedir(dir);
	return count;
}

int send_waiting()
{
	log = stdout;
	read_config();
	printf("Sending pending messages:\n"); fflush(stdout);
	return for_every(sendmes);
}

void show_waiting()
{
	printf("Pending messages:\n");
	for_every(show);
}

int main(int argc, char **argv)
{
	if (argc==1)
		show_waiting();
	else
	{
		if (strcmp(argv[1], "d")==0)
			deamon();
		else
		if (strcmp(argv[1], "s")==0)
			send_waiting();
		else
			printf("Illegal option\n");
	}
	return 0;
}
